import pyspark.sql.functions as F
from pyspark.sql.column import Column
from laktory.spark.functions._common import (
    COLUMN_OR_NAME,
    INT_OR_COLUMN,
    FLOAT_OR_COLUMN,
    STRING_OR_COLUMN,
    _col,
    _lit,
)

__all__ = [
    "add",
    "sub",
    "mul",
    "div",
    "poly1",
    "poly2",
    "scaled_power",
    "roundp",
]


# --------------------------------------------------------------------------- #
# Arithmetics                                                                 #
# --------------------------------------------------------------------------- #


def add(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Get floating addition `x + a`

    Parameters
    ----------
    x:
        Input column
    a:
        Addend

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[8], [6]], ["x"])
    df = df.withColumn("y", LF.poly1("x", a=2))
    print(df.show_string())
    '''
    +---+----+
    |  x|   y|
    +---+----+
    |  8|16.0|
    |  6|12.0|
    +---+----+
    '''
    ```
    """
    return _col(x) + _lit(a)


def sub(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Get floating subtraction `x - a`

    Parameters
    ----------
    x:
        Input column
    a:
        Subtrahend

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[8], [6]], ["x"])
    df = df.withColumn("y", LF.poly1("x", a=2))
    print(df.show_string())
    '''
    +---+----+
    |  x|   y|
    +---+----+
    |  8|16.0|
    |  6|12.0|
    +---+----+
    '''
    ```
    """
    return _col(x) - _lit(a)


def mul(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Get floating multiplication `x * a`

    Parameters
    ----------
    x:
        Input column
    a:
        Multiplier

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[8], [6]], ["x"])
    df = df.withColumn("y", LF.poly1("x", a=2))
    print(df.show_string())
    '''
    +---+----+
    |  x|   y|
    +---+----+
    |  8|16.0|
    |  6|12.0|
    +---+----+
    '''
    ```
    """
    return _col(x) * _lit(a)


def div(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Get floating division `x / a`

    Parameters
    ----------
    x:
        Input column
    a:
        Divider

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[8], [6]], ["x"])
    df = df.withColumn("y", LF.poly1("x", a=2))
    print(df.show_string())
    '''
    +---+----+
    |  x|   y|
    +---+----+
    |  8|16.0|
    |  6|12.0|
    +---+----+
    '''
    ```
    """
    return _col(x) / _lit(a)


# --------------------------------------------------------------------------- #
# Polynomials                                                                 #
# --------------------------------------------------------------------------- #


def poly1(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    b: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Polynomial function of first degree `a * x + b`

    Parameters
    ----------
    x:
        Input column
    a:
        Slope
    b:
        y-intercept

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory  # noqa: F401
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[9]], ["x"])
    df = df.withColumn("y", LF.poly1("x", a=-1, b=2))
    print(df.show_string())
    '''
    +---+---+
    |  x|  y|
    +---+---+
    |  9| -7|
    +---+---+
    '''
    ```
    """
    return _lit(a) * _col(x) + _lit(b)


def poly2(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    b: FLOAT_OR_COLUMN = 0.0,
    c: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Polynomial function of second degree `a * x**2 + b * x + c`

    Parameters
    ------
    x:
        Input column
    a:
        x**2 coefficient
    b:
        x**1 coefficient
    c:
        x**0 coefficient

    Returns
    -------
    :
        Output column


    Examples
    --------
    ```py
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[9]], ["x"])
    df = df.withColumn("y", LF.poly2("x", a=-1, b=2))
    print(df.show_string())
    '''
    +---+-----+
    |  x|    y|
    +---+-----+
    |  9|-63.0|
    +---+-----+
    '''
    ```
    """
    return _lit(a) * _col(x) ** 2 + _lit(b) * _col(x) + _lit(c)


# --------------------------------------------------------------------------- #
# Scaled Power                                                                #
# --------------------------------------------------------------------------- #


def scaled_power(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    n: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Scaled power function `a * x`

    Parameters
    ------
    x:
        Input column
    a:
        Scaling coefficient
    n:
        Exponent

    Returns
    -------
    :
        Output column


    Examples
    --------
    ```py
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[9]], ["x"])
    df = df.withColumn("y", LF.scaled_power("x", a=-3, n=2))
    print(df.show_string())
    '''
    +---+------+
    |  x|     y|
    +---+------+
    |  9|-243.0|
    +---+------+
    '''
    ```
    """
    return _lit(a) * _col(x) ** _lit(n)


# --------------------------------------------------------------------------- #
# Rounding                                                                    #
# --------------------------------------------------------------------------- #


def roundp(
    x: COLUMN_OR_NAME,
    p: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Evenly round to the given precision

    Parameters
    ------
    x:
        Input column
    p:
        Precision

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[0.781], [13.0]], ["x"])
    df = df.withColumn("y", LF.roundp("x", p=5))
    print(df.show_string())
    '''
    +-----+----+
    |    x|   y|
    +-----+----+
    |0.781| 0.0|
    | 13.0|15.0|
    +-----+----+
    '''

    df = df.withColumn("y", LF.roundp("x", p=0.25))
    print(df.show_string())
    '''
    +-----+----+
    |    x|   y|
    +-----+----+
    |0.781|0.75|
    | 13.0|13.0|
    +-----+----+
    '''
    ```
    """
    # eps0 = 1.0e-16
    # precision = float(precision)
    # if precision < eps0:
    #     raise ValueError("Precision must be greater than 1.0e-16")
    return F.round(_col(x) / _lit(p)) * _lit(p)



import typer

APP_NAME = "laktory-cli"

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    help="Laktory CLI to preview and deploy resources",
)  # prevent display secret data


if __name__ == "__main__":
    app()



import os
import laktory

from laktory._useragent import DATABRICKS_USER_AGENT
from laktory._version import VERSION


def test_user_agent():
    assert os.getenv("DATABRICKS_SDK_UPSTREAM") == DATABRICKS_USER_AGENT
    assert os.getenv("DATABRICKS_SDK_UPSTREAM_VERSION") == VERSION


if __name__ == "__main__":
    test_user_agent()



from pydantic import field_validator
from typing import Union

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_TYPES
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class Column(BaseModel):
    """
    Definition of a table column, including instructions on how to build the
    column using spark and an input dataframe.

    Attributes
    ----------
    catalog_name:
        Name of the catalog string the column table
    comment:
        Text description of the column
    name:
        Name of the column
    pii:
        If `True`, the column is flagged as Personally Identifiable Information
    schema_name:
        Name of the schema storing the column table
    table_name:
        Name of the table storing the column.
    type:
        Column data type
    unit:
        Column units
    """

    catalog_name: Union[str, None] = None
    comment: Union[str, None] = None
    name: str
    pii: Union[bool, None] = None
    raise_missing_arg_exception: Union[bool, None] = True
    schema_name: Union[str, None] = None
    table_name: Union[str, None] = None
    type: str = "string"
    unit: Union[str, None] = None

    @field_validator("type")
    def default_load_path(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_TYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_TYPES}"
                )
        return v

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        if self.table_name:
            if _id == "":
                _id = self.table_name
            else:
                _id += f".{self.table_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Column full name `{catalog_name}.{schema_name}.{table_name}.{column_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id



import os
import json
import yaml

from laktory.models import Column
from laktory._testing import Paths
from laktory._testing.stockprices import df_brz as df

paths = Paths(__file__)
GOOGL = {
    "name": "close",
    "type": "double",
    "unit": "USD",
    "catalog_name": "dev",
    "schema_name": "markets",
    "table_name": "stock_prices",
}

with open(os.path.join(paths.tmp, "googl.json"), "w") as fp:
    json.dump(GOOGL, fp)
with open(os.path.join(paths.tmp, "googl.yaml"), "w") as fp:
    yaml.dump(GOOGL, fp)

# Spark
df = df.filter("data.symbol == 'GOOGL'").limit(5)


def test_model():
    c0 = Column(**GOOGL)
    c1 = Column.model_validate(GOOGL)
    assert c1.type == "double"
    assert c1.catalog_name == "dev"
    assert c1.schema_name == "markets"
    assert c1.table_name == "stock_prices"
    assert c1.full_name == "dev.markets.stock_prices.close"
    assert c0 == c1


def test_read():
    c0 = Column(**GOOGL)

    with open(f"{paths.tmp}/googl.yaml", "r") as fp:
        c1 = Column.model_validate_yaml(fp)

    with open(f"{paths.tmp}/googl.json", "r") as fp:
        c2 = Column.model_validate_json_file(fp)

    assert c1 == c0
    assert c2 == c0


if __name__ == "__main__":
    test_model()
    test_read()



import os
import subprocess
from pydantic import BaseModel
from typing import Union
from prompt_toolkit.validation import Validator
from prompt_toolkit.validation import ValidationError

from laktory.models.stacks.stack import Stack
from laktory.constants import SUPPORTED_BACKENDS
from laktory._logger import get_logger

logger = get_logger(__name__)
DIRPATH = os.path.dirname(__file__)


class BackendValidator(Validator):
    def validate(self, document):
        text = document.text
        if text.lower() not in SUPPORTED_BACKENDS:
            raise ValidationError(
                message=f"Please enter one of the supported IaC backends {SUPPORTED_BACKENDS}",
                cursor_position=len(text),
            )  # Move cursor to end


class CLIController(BaseModel):
    stack_filepath: Union[str, None] = None
    backend: Union[str, None] = None
    organization: Union[str, None] = None
    env: Union[str, None] = None
    auto_approve: Union[bool, None] = False
    pulumi_options_str: Union[str, None] = None
    terraform_options_str: Union[str, None] = None
    stack: Union[Stack, None] = None

    def model_post_init(self, __context):
        super().model_post_init(__context)

        # Read stack
        if self.stack_filepath is None:
            self.stack_filepath = "./stack.yaml"
        with open(self.stack_filepath, "r") as fp:
            self.stack = Stack.model_validate_yaml(fp)

        # Set backend
        if self.backend is None:
            self.backend = self.stack.backend
        if self.backend is None:
            raise ValueError(
                "backend ['pulumi', 'terraform'] must be specified in stack file or as CLI option "
            )

        # Set organization
        if self.organization is None:
            self.organization = self.stack.organization
        if self.organization is None and self.backend == "pulumi":
            raise ValueError("organization must be specified with pulumi backend")

        # Check environment
        if self.env is None:
            env_names = list(self.stack.envs.keys())
            if env_names:
                logger.warn(
                    f"Environment not specified, defaulting to first available ({env_names[0]})"
                )
                self.env = env_names[0]

    @property
    def pulumi_options(self):
        options = []
        if self.auto_approve:
            options += ["--yes"]

        if self.pulumi_options_str:
            options += self.pulumi_options_str.split(",")

        return options

    @property
    def terraform_options(self):
        options = []
        if self.auto_approve:
            options += ["-auto-approve"]

        if self.terraform_options_str:
            options = self.terraform_options_str.split(",")

        return options

    @property
    def pulumi_stack_name(self):
        return self.organization + "/" + self.env

    def pulumi_call(self, cmd):
        if self.pulumi_stack_name is None:
            raise ValueError("Argument `stack` must be specified with pulumi backend")

        pstack = self.stack.to_pulumi(env=self.env)
        getattr(pstack, cmd)(stack=self.pulumi_stack_name, flags=self.pulumi_options)

    def terraform_call(self, cmd):
        pstack = self.stack.to_terraform(env=self.env)
        getattr(pstack, cmd)(flags=self.terraform_options)


class Worker:
    def run(self, cmd, cwd=None, raise_exceptions=True):
        try:
            completed_process = subprocess.run(
                cmd,
                cwd=cwd,
                check=True,
            )

        except Exception as e:
            if raise_exceptions:
                raise e
            else:
                print("An error occurred:", str(e))



import pandas as pd
from pyspark.sql import SparkSession
import pyspark.sql.functions as F
import pyspark.sql.types as T
import os
import yfinance as yf

from laktory import models
from datetime import datetime

# --------------------------------------------------------------------------- #
# Setup                                                                       #
# --------------------------------------------------------------------------- #

spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
spark.conf.set("spark.sql.session.timeZone", "UTC")

symbols = [
    "AAPL",
    "AMZN",
    "GOOGL",
    "MSFT",
]

rootpath = os.path.join(os.path.dirname(__file__), "data/")

t0 = datetime(2023, 9, 1)
t1 = datetime(2023, 9, 30)


# --------------------------------------------------------------------------- #
# Build Events                                                                #
# --------------------------------------------------------------------------- #


events = []
for s in symbols:
    df = yf.download(s, t0, t1, interval="1d")
    for _, row in df.iterrows():
        events += [
            models.DataEvent(
                name="stock_price",
                producer={
                    "name": "yahoo-finance",
                },
                data={
                    "created_at": _,
                    "symbol": s,
                    "open": float(
                        row["Open"]
                    ),  # np.float64 are not supported for serialization
                    "close": float(row["Close"]),
                    "high": float(row["High"]),
                    "low": float(row["Low"]),
                    "@id": "_id",
                },
            )
        ]
        events[-1].events_root_ = os.path.join(rootpath, "events/")


# --------------------------------------------------------------------------- #
# Write Events                                                                #
# --------------------------------------------------------------------------- #

for event in events:
    event.to_path(suffix=event.data["symbol"], skip_if_exists=True)


# --------------------------------------------------------------------------- #
# Build Pandas DataFrame                                                      #
# --------------------------------------------------------------------------- #

pdf = pd.DataFrame([e.model_dump() for e in events])
del pdf["event_root"]
# pdf.to_json(os.path.join(rootpath, "brz_stock_prices.json"))


# --------------------------------------------------------------------------- #
# Write Bronze DataFrame                                                      #
# --------------------------------------------------------------------------- #

df = spark.createDataFrame(
    pdf,
    schema=T.StructType(
        [
            T.StructField("name", T.StringType()),
            T.StructField("description", T.StringType()),
            T.StructField(
                "producer",
                T.StructType(
                    [
                        T.StructField("name", T.StringType()),
                        T.StructField("description", T.StringType()),
                        T.StructField("party", T.IntegerType()),
                    ]
                ),
            ),
            T.StructField(
                "data",
                T.StructType(
                    [
                        T.StructField("created_at", T.StringType()),
                        T.StructField("symbol", T.StringType()),
                        T.StructField("open", T.DoubleType()),
                        T.StructField("close", T.DoubleType()),
                        T.StructField("high", T.DoubleType()),
                        T.StructField("low", T.DoubleType()),
                        T.StructField("_name", T.StringType()),
                        T.StructField("_producer_name", T.StringType()),
                        T.StructField("_created_at", T.StringType()),
                        T.StructField("@id", T.StringType()),
                    ]
                ),
            ),
        ]
    ),
)
df = df.repartition(1)
df.write.parquet(os.path.join(rootpath, "brz_stock_prices"), mode="overwrite")


# --------------------------------------------------------------------------- #
# Build Silver DataFrame                                                      #
# --------------------------------------------------------------------------- #

# Stock prices
cols0 = df.columns
df = df.withColumn("created_at", F.col("data._created_at").cast(T.TimestampType()))
df = df.withColumn("symbol", F.col("data.symbol").cast(T.StringType()))
df = df.withColumn("open", F.col("data.open").cast(T.DoubleType()))
df = df.withColumn("close", F.col("data.close").cast(T.DoubleType()))
df = df.drop(*cols0)
df = df.repartition(1)
df.write.parquet(os.path.join(rootpath, "slv_stock_prices"), mode="overwrite")

# Metadata
df_meta = spark.createDataFrame(
    pd.DataFrame(
        {
            "symbol2": ["AAPL", "GOOGL", "AMZN"],
            "currency": ["USD"] * 3,
            "first_traded": [
                "1980-12-12T14:30:00.000Z",
                "2004-08-19T13:30:00.00Z",
                "1997-05-15T13:30:00.000Z",
            ],
        }
    )
)
df_meta = df_meta.repartition(1)
df_meta.write.format("parquet").save(
    os.path.join(rootpath, "slv_stock_meta"), mode="overwrite"
)



import os

from .stockprices import df_brz
from .stockprices import df_slv
from .stockprices import df_meta
from .stockprices import spark
from .stockprices import table_slv
from .stockprices import table_slv_join


class Paths:
    def __init__(self, file):
        self.root = os.path.dirname(file)
        self.data = os.path.join(self.root, "data")
        self.tmp = os.path.join(self.root, "tmp")

        if not os.path.exists(self.tmp):
            os.makedirs(self.tmp)



import os
import pandas as pd
from pyspark.sql import SparkSession

from laktory.models import Table

spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
spark.conf.set("spark.sql.session.timeZone", "UTC")


# --------------------------------------------------------------------------- #
# DataFrames                                                                  #
# --------------------------------------------------------------------------- #

data_dirpath = os.path.join(os.path.dirname(__file__), "../../tests/data/")
df_brz = spark.read.parquet(os.path.join(data_dirpath, "brz_stock_prices"))
df_slv = spark.read.parquet(os.path.join(data_dirpath, "slv_stock_prices"))
df_meta = spark.read.parquet(os.path.join(data_dirpath, "slv_stock_meta"))
df_name = spark.createDataFrame(
    pd.DataFrame(
        {
            "symbol3": ["AAPL", "GOOGL", "AMZN"],
            "name": ["Apple", "Google", "Amazon"],
        }
    )
)

# --------------------------------------------------------------------------- #
# Tables                                                                      #
# --------------------------------------------------------------------------- #

table_slv = Table(
    name="slv_stock_prices",
    columns=[
        {
            "name": "created_at",
            "type": "timestamp",
        },
        {
            "name": "symbol",
            "type": "string",
        },
        {
            "name": "open",
            "type": "double",
        },
        {
            "name": "close",
            "type": "double",
        },
    ],
    data=[
        ["2023-11-01T00:00:00Z", "AAPL", 1, 2],
        ["2023-11-01T01:00:00Z", "AAPL", 3, 4],
        ["2023-11-01T00:00:00Z", "GOOGL", 3, 4],
        ["2023-11-01T01:00:00Z", "GOOGL", 5, 6],
    ],
    catalog_name="dev",
    schema_name="markets",
    builder={
        "table_source": {
            "name": "brz_stock_prices",
        },
        "layer": "SILVER",
    },
    expectations=[
        {"name": "positive_price", "expression": "open > 0", "action": "FAIL"},
        {
            "name": "recent_price",
            "expression": "created_at > '2023-01-01'",
            "action": "DROP",
        },
    ],
)

table_slv_join = Table(
    name="slv_join_stock_prices",
    catalog_name="dev",
    schema_name="markets",
    builder={
        "layer": "SILVER",
        "table_source": {
            # "mock_df": df_slv,
            "name": "slv_stock_prices",
            "filter": "created_at = '2023-09-01T00:00:00Z'",
        },
        "spark_chain": {
            "nodes": [
                {
                    "spark_func_name": "smart_join",
                    "spark_func_kwargs": {
                        "other": {
                            "name": "slv_stockmeta",
                            # "mock_df": df_meta,
                            "renames": {"symbol2": "symbol"},
                        },
                        "on": ["symbol"],
                    },
                },
                {"column": {"name": "symbol3"}, "sql_expression": "symbol"},
                {
                    "spark_func_name": "drop",
                    "spark_func_args": ["symbol"],
                },
                {
                    "spark_func_name": "smart_join",
                    "spark_func_kwargs": {
                        "other": {
                            "name": "slv_stock_names",
                            # "mock_df": df_name,
                        },
                        "on": ["symbol3"],
                    },
                },
            ],
        },
        "drop_source_columns": False,
    },
)

# Because a pipeline will modify the schema and other properties of the table
# we need to be a copy specifically for the pipeline. Deep copy does not work
# so well on spark dataframes. They are assigned after the copy.
table_slv_pl = table_slv.model_copy(deep=True)
table_slv_join_pl = table_slv_join.model_copy(deep=True)
table_slv_join.builder.table_source.mock_df = df_slv
table_slv_join.builder.spark_chain.nodes[0].spark_func_kwargs[
    "other"
].value.mock_df = df_meta
table_slv_join.builder.spark_chain.nodes[3].spark_func_kwargs[
    "other"
].value.mock_df = df_name



import pyspark.sql.functions as F

from pyspark.sql.column import Column
from laktory.spark.functions._common import (
    COLUMN_OR_NAME,
    INT_OR_COLUMN,
    FLOAT_OR_COLUMN,
    STRING_OR_COLUMN,
    _col,
    _lit,
)

__all__ = [
    "string_split",
    "uuid",
]


# --------------------------------------------------------------------------- #
# string_split                                                                #
# --------------------------------------------------------------------------- #


def string_split(
    x: COLUMN_OR_NAME,
    pattern: str,
    key: int,
) -> Column:
    """
    Get substring using separator `pat`.

    Parameters
    ----------
    x:
        Input text series to split
    pattern:
        String or regular expression to split on. If not specified, split on whitespace.
    key:
        Split index to return

    Returns
    -------
    :
        Result

    Examples
    --------
    ```py
    from pyspark.sql import SparkSession
    import pyspark.sql.functions as F
    import laktory.spark.functions as LF

    spark = SparkSession.builder.getOrCreate()

    df = spark.range(1).withColumn("x", F.lit("price_close"))
    df = df.withColumn("y", LF.string_split("x", pattern="_", key=1))
    print(df.show_string())
    '''
    +---+-----------+-----+
    | id|          x|    y|
    +---+-----------+-----+
    |  0|price_close|close|
    +---+-----------+-----+
    '''
    ```
    """
    return F.split(_col(x), pattern=pattern).getItem(key)


# --------------------------------------------------------------------------- #
# uuid                                                                        #
# --------------------------------------------------------------------------- #


def uuid() -> Column:
    """
    Create a unique id for each row.

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import laktory.spark.functions as LF

    df = spark.range(3)
    df = df.withColumn("uuid", LF.uuid())
    '''
    +---+--------------------+
    | id|                uuid|
    +---+--------------------+
    |  0|acc0b53e-a36f-4f8...|
    |  1|56cdeb41-6828-486...|
    |  2|64a7d2bf-5e1d-41a...|
    +---+--------------------+
    '''
    ```
    """
    return F.expr("uuid()")



import time
from typing import Literal
from databricks.sdk.service.pipelines import StartUpdateResponse
from databricks.sdk.service.pipelines import GetUpdateResponse
from databricks.sdk.service.pipelines import UpdateInfoState
from databricks.sdk.service.pipelines import EventLevel
from databricks.sdk.core import DatabricksError

from laktory.dispatcher.dispatcherrunner import DispatcherRunner
from laktory.datetime import unix_timestamp
from laktory._logger import get_logger

logger = get_logger(__name__)


class PipelineRunner(DispatcherRunner):
    """
    Pipeline runner.
    """

    _update_start: StartUpdateResponse = None
    _update: GetUpdateResponse = None

    def get_id(self) -> str:
        """Get deployed pipeline id"""
        logger.info(f"Getting id for pipeline {self.name}")
        for _pl in self.wc.pipelines.list_pipelines(
            filter=f"name LIKE '{self.name}'", max_results=1
        ):
            if self.name == _pl.name:
                self.id = _pl.pipeline_id
                return self.id

    def run(
        self,
        wait: bool = True,
        timeout: int = 20 * 60,
        full_refresh: bool = False,
        raise_exception: bool = False,
        current_run_action: Literal["WAIT", "CANCEL", "FAIL"] = "WAIT",
    ):
        """
        Run remote pipeline and monitor failures.

        Parameters
        ----------
        wait:
            If `True`, runs synchronously and wait for completion.
        timeout:
            Maximum time allowed for the pipeline to run before an exception is
            raised.
        full_refresh:
            If `True`, tables are fully refreshed (re-built). Otherwise, only
            increments are processed.
        raise_exception:
            If `True`, exceptions are raised if the pipeline fails.
        current_run_action:
            Action to take with respect to current run (if any). Possible
            options are:
                - WAIT: wait for the current run to complete
                - CANCEL: cancel the current run
                - FAIL: raise an exception

        Returns
        -------
        output:
            None
        """
        event_ids = []

        # Start update
        t0 = time.time()
        try:
            self._update_start = self.wc.pipelines.start_update(
                pipeline_id=self.id,
                full_refresh=full_refresh,
                # validate_only=False,
            )
            logger.info(f"Pipeline {self.name} update started...")

        except DatabricksError as e:
            # Exception is raised when update is currently in progress
            if current_run_action.upper() == "FAIL":
                logger.info(f"Pipeline {self.name} already running...")
                raise e

            elif current_run_action.upper() == "WAIT":
                logger.info(
                    f"Pipeline {self.name} waiting for current update to be completed..."
                )
                self.wc.pipelines.wait_get_pipeline_idle(pipeline_id=self.id)

            elif current_run_action.upper() == "CANCEL":
                logger.info(f"Pipeline {self.name} cancelling current update...")
                self.wc.pipelines.stop_and_wait(pipeline_id=self.id)

            logger.info(f"Pipeline {self.name} update started...")
            self._update_start = self.wc.pipelines.start_update(
                pipeline_id=self.id,
                full_refresh=full_refresh,
                # validate_only=False,
            )

        pstate = None
        self.get_update()
        logger.info(f"Pipeline {self.name} run URL: {self.update_url}")
        if wait:
            while (
                time.time() - t0 < timeout
                or self.update_state == UpdateInfoState.COMPLETED
            ):
                self.get_update()

                if self.update_state != pstate:
                    logger.info(
                        f"Pipeline {self.name} state: {self.update_state.value}"
                    )
                    pstate = self.update_state

                # filter don't seem to work
                for event in self.wc.pipelines.list_pipeline_events(
                    pipeline_id=self.id,
                    max_results=100,
                    # filter=f"(level in ('ERROR', 'WARN')) AND (timestamp > '{utc_datetime(t0).isoformat()}Z')"
                    # filter=f"level in ('ERROR', 'WARN')",
                    # filter=f"timestamp > '{utc_datetime(t0).isoformat()}Z'"
                    page_token=None,
                ):
                    if (
                        event.id in event_ids
                        or unix_timestamp(event.timestamp) < t0
                        or event.origin.update_id != self.update_id
                    ):
                        continue

                    if event.level == EventLevel.WARN:
                        logger.warn(event.message)
                    elif event.level == EventLevel.ERROR:
                        logger.error(event.message)
                    event_ids += [event.id]

                if self.update_state in [
                    UpdateInfoState.CANCELED,
                    UpdateInfoState.COMPLETED,
                    UpdateInfoState.FAILED,
                ]:
                    break

                time.sleep(1.0)

            logger.info(
                f"Pipeline {self.name} update terminated after {time.time() - t0: 5.2f} sec with {self.update_state}"
            )
            if raise_exception and self.update_state != UpdateInfoState.COMPLETED:
                raise Exception(
                    f"Pipeline {self.name} update not completed ({self.update_state})"
                )

    def get_update(self):
        self._update = self.wc.pipelines.get_update(
            pipeline_id=self.id, update_id=self.update_id
        )
        return self._update

    @property
    def update_state(self):
        return self._update.update.state

    @property
    def update_id(self) -> str:
        return self._update_start.update_id

    @property
    def update_url(self) -> str:
        return f"{self.wc.config.host}/pipelines/{self.id}/updates/{self.update_id}"



from pydantic import Field
from laktory.models.providers.baseprovider import BaseProvider
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class DatabricksProvider(BaseProvider, PulumiResource, TerraformResource):
    """
    Databricks Provider

    Attributes
    ----------
    account_id:
        Account Id that could be found in the bottom left corner of Accounts
        Console.
    auth_type:
        Enforce specific auth type to be used in very rare cases, where a
        single provider state manages Databricks workspaces on more than one
        cloud and More than one authorization method configured error is a
        false positive. Valid values are pat, basic, azure-client-secret,
        azure-msi, azure-cli, and databricks-cli.
    azure_client_id:
        TODO
    azure_client_secret:
        TODO
    azure_environment:
        TODO
    azure_login_app_id:
        TODO
    azure_tenant_id:
        TODO
    azure_use_msi:
        TODO
    azure_workspace_resource_id:
        TODO
    client_id:
        TODO
    client_secret:
        TODO
    cluster_id:
        TODO
    config_file:
        Location of the Databricks CLI credentials file created by databricks
         configure --token command (~/.databrickscfg by default). Check
         Databricks CLI documentation for more details. The provider uses
          configuration file credentials when you don't specify
          host/token/username/password/azure attributes.
    databricks_cli_path:
        TODO
    debug_headers:
        TODO
    debug_truncate_bytes:
        TODO
    google_credentials:
        TODO
    google_service_account:
        TODO
    host:
        This is the host of the Databricks workspace. It is a URL that you use
        to login to your workspace.
    http_timeout_seconds:
        TODO
    metadata_service_url:
        TODO
    password:
        This is the user's password that can log into the workspace.
    profile:
        Connection profile specified within ~/.databrickscfg. Please check
         connection profiles section for more details.
    rate_limit:
        TODO
    retry_timeout_seconds:
        TODO
    skip_verify:
        TODO
    token:
        This is the API token to authenticate into the workspace.
    username:
        This is the username of the user that can log into the workspace.
    warehouse_id:
        TODO

    Examples
    --------
    ```py
    from laktory import models

    p = models.DatabricksProvider(
        host="adb-4623853922539974.14.azuredatabricks.net",
        token="${vars.DATABRICKS_TOKEN}",
    )
    ```
    """

    source: str = Field("databricks/databricks", exclude=True)
    version: str = Field(">=1.39", exclude=True)  # Required to monitor laktory usage

    account_id: str = None
    auth_type: str = None
    azure_client_id: str = None
    azure_client_secret: str = None
    azure_environment: str = None
    azure_login_app_id: str = None
    azure_tenant_id: str = None
    azure_use_msi: bool = None
    azure_workspace_resource_id: str = None
    client_id: str = None
    client_secret: str = None
    cluster_id: str = None
    config_file: str = None
    databricks_cli_path: str = None
    debug_headers: bool = None
    debug_truncate_bytes: int = None
    google_credentials: str = None
    google_service_account: str = None
    host: str = None
    http_timeout_seconds: int = None
    metadata_service_url: str = None
    password: str = None
    profile: str = None
    rate_limit: int = None
    retry_timeout_seconds: int = None
    skip_verify: bool = None
    token: str = None
    username: str = None
    warehouse_id: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # @property
    # def resource_key(self) -> str:
    #     return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "pulumi:providers:databricks"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Provider
